@@ -23,6 +23,7 @@
 #include <gtk/gtk.h>
 #include "ev-document.h"
 #include "ev-window.h"
+#include "ev-file-exporter.h"
 
 G_BEGIN_DECLS
 
@@ -119,6 +120,7 @@ struct _EvJobRender
 	EvJob parent;
 
 	EvRenderContext *rc;
+	gboolean page_ready;
 	gint target_width;
 	gint target_height;
 	cairo_surface_t *surface;
@@ -144,6 +146,8 @@ struct _EvJobRender
 struct _EvJobRenderClass
 {
 	EvJobClass parent_class;
+
+	void    (* page_ready) (EvJobRender *job);
 };
 
 struct _EvJobThumbnail
@@ -196,10 +200,12 @@ struct _EvJobPrint
 	gint n_ranges;
 	EvPrintPageSet page_set;
 	gint copies;
+	gint pages_per_sheet;
 	gboolean collate;
 	gboolean reverse;
 	gdouble width;
 	gdouble height;
+	EvFileExporterOrientation orientation;
 };
 
 struct _EvJobPrintClass
@@ -258,9 +264,11 @@ EvJob          *ev_job_print_new          (EvDocument      *document,
 					   const gchar     *format,
 					   gdouble          width,
 					   gdouble          height,
+					   EvFileExporterOrientation orientation,
 					   EvPrintRange    *ranges,
 					   gint             n_ranges,
 					   EvPrintPageSet   page_set,
+					   gint             pages_per_sheet,
 					   gint             copies,
 					   gdouble          collate,
 					   gdouble          reverse);
