@@ -3,17 +3,24 @@
 #include "ev-page-cache.h"
 #include "ev-selection.h"
 #include "ev-document-images.h"
+#include "ev-document-forms.h"
 #include "ev-image.h"
+#include "ev-form-field.h"
 
 typedef struct _CacheJobInfo
 {
 	EvJob *job;
 	EvRenderContext *rc;
+	gboolean page_ready;
+
+	/* Region of the page that needs to be drawn */
+	GdkRegion *region; 
 
 	/* Data we get from rendering */
 	cairo_surface_t *surface;
 	GList *link_mapping;
 	GList *image_mapping;
+	GList *form_field_mapping;
 	GdkRegion *text_mapping;
 	
 	/* Selection data. 
@@ -67,6 +74,8 @@ static void          ev_pixbuf_cache_init       (EvPixbufCache      *pixbuf_cach
 static void          ev_pixbuf_cache_class_init (EvPixbufCacheClass *pixbuf_cache);
 static void          ev_pixbuf_cache_finalize   (GObject            *object);
 static void          ev_pixbuf_cache_dispose    (GObject            *object);
+static void          job_page_ready_cb          (EvJob              *job,
+						 EvPixbufCache      *pixbuf_cache);
 static void          job_finished_cb            (EvJob              *job,
 						 EvPixbufCache      *pixbuf_cache);
 static CacheJobInfo *find_job_cache             (EvPixbufCache      *pixbuf_cache,
@@ -74,6 +83,9 @@ static CacheJobInfo *find_job_cache             (EvPixbufCache      *pixbuf_cach
 static void          copy_job_to_job_info       (EvJobRender        *job_render,
 						 CacheJobInfo       *job_info,
 						 EvPixbufCache      *pixbuf_cache);
+static void          copy_job_page_and_selection_to_job_info (EvJobRender        *job_render,
+							      CacheJobInfo       *job_info,
+							      EvPixbufCache      *pixbuf_cache);
 static gboolean      new_selection_surface_needed(EvPixbufCache      *pixbuf_cache,
 						  CacheJobInfo       *job_info,
 						  gint                page,
@@ -112,13 +124,15 @@ ev_pixbuf_cache_class_init (EvPixbufCacheClass *class)
 	object_class->finalize = ev_pixbuf_cache_finalize;
 	object_class->dispose = ev_pixbuf_cache_dispose;
 
-	signals[JOB_FINISHED] = g_signal_new ("job-finished",
-					    G_OBJECT_CLASS_TYPE (object_class),
-					    G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
-					    G_STRUCT_OFFSET (EvPixbufCacheClass, job_finished),
-					    NULL, NULL,
-					    g_cclosure_marshal_VOID__VOID,
-					    G_TYPE_NONE, 0);
+	signals[JOB_FINISHED] =
+		g_signal_new ("job-finished",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
+			      G_STRUCT_OFFSET (EvPixbufCacheClass, job_finished),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__POINTER,
+			      G_TYPE_NONE, 1,
+			      G_TYPE_POINTER);
 }
 
 static void
@@ -142,6 +156,9 @@ dispose_cache_job_info (CacheJobInfo *job_info,
 	if (job_info == NULL)
 		return;
 	if (job_info->job) {
+		g_signal_handlers_disconnect_by_func (job_info->job,
+						      G_CALLBACK (job_page_ready_cb),
+						      data);
 		g_signal_handlers_disconnect_by_func (job_info->job,
 						      G_CALLBACK (job_finished_cb),
 						      data);
@@ -153,6 +170,10 @@ dispose_cache_job_info (CacheJobInfo *job_info,
 		cairo_surface_destroy (job_info->surface);
 		job_info->surface = NULL;
 	}
+	if (job_info->region) {
+		gdk_region_destroy (job_info->region);
+		job_info->region = NULL;
+	}
 	if (job_info->link_mapping) {
 		ev_link_mapping_free (job_info->link_mapping);
 		job_info->link_mapping = NULL;
@@ -161,6 +182,10 @@ dispose_cache_job_info (CacheJobInfo *job_info,
 		ev_image_mapping_free (job_info->image_mapping);
 		job_info->image_mapping = NULL;
 	}
+	if (job_info->form_field_mapping) {
+		ev_form_field_mapping_free (job_info->form_field_mapping);
+		job_info->form_field_mapping = NULL;
+	}
 	if (job_info->text_mapping) {
 		gdk_region_destroy (job_info->text_mapping);
 		job_info->text_mapping = NULL;
@@ -216,6 +241,26 @@ ev_pixbuf_cache_new (GtkWidget  *view,
 	return pixbuf_cache;
 }
 
+static void
+job_page_ready_cb (EvJob         *job,
+		   EvPixbufCache *pixbuf_cache)
+{
+	CacheJobInfo *job_info;
+	EvJobRender *job_render = EV_JOB_RENDER (job);
+
+	/* If the job is outside of our interest, we silently discard it */
+	if ((job_render->rc->page < (pixbuf_cache->start_page - pixbuf_cache->preload_cache_size)) ||
+	    (job_render->rc->page > (pixbuf_cache->end_page + pixbuf_cache->preload_cache_size))) {
+		g_object_unref (job);
+		return;
+	}
+
+	job_info = find_job_cache (pixbuf_cache, job_render->rc->page);
+
+	copy_job_page_and_selection_to_job_info (job_render, job_info, pixbuf_cache);
+	g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0, job_info->region);
+}
+
 static void
 job_finished_cb (EvJob         *job,
 		 EvPixbufCache *pixbuf_cache)
@@ -229,11 +274,9 @@ job_finished_cb (EvJob         *job,
 		g_object_unref (job);
 		return;
 	}
-	
-	job_info = find_job_cache (pixbuf_cache, job_render->rc->page);
 
+	job_info = find_job_cache (pixbuf_cache, job_render->rc->page);
 	copy_job_to_job_info (job_render, job_info, pixbuf_cache);
-	g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
 }
 
 /* This checks a job to see if the job would generate the right sized pixbuf
@@ -320,17 +363,17 @@ move_one_job (CacheJobInfo  *job_info,
 
 	*target_page = *job_info;
 	job_info->job = NULL;
+	job_info->region = NULL;
 	job_info->surface = NULL;
 	job_info->link_mapping = NULL;
 	job_info->image_mapping = NULL;
+	job_info->form_field_mapping = NULL;
 
 	if (new_priority != priority && target_page->job) {
 		ev_job_queue_update_job (target_page->job, new_priority);
 	}
 }
 
-
-
 static void
 ev_pixbuf_cache_update_range (EvPixbufCache *pixbuf_cache,
 			      gint           start_page,
@@ -403,13 +446,10 @@ ev_pixbuf_cache_update_range (EvPixbufCache *pixbuf_cache,
 }
 
 static void
-copy_job_to_job_info (EvJobRender   *job_render,
-		      CacheJobInfo  *job_info,
-		      EvPixbufCache *pixbuf_cache)
+copy_job_page_and_selection_to_job_info (EvJobRender   *job_render,
+					 CacheJobInfo  *job_info,
+					 EvPixbufCache *pixbuf_cache)
 {
-
-	job_info->points_set = FALSE;
-
 	if (job_info->surface) {
 		cairo_surface_destroy (job_info->surface);
 	}
@@ -420,24 +460,7 @@ copy_job_to_job_info (EvJobRender   *job_render,
 	}
 	job_info->rc = g_object_ref (job_render->rc);
 
-	if (job_render->include_links) {
-		if (job_info->link_mapping)
-			ev_link_mapping_free (job_info->link_mapping);
-		job_info->link_mapping = job_render->link_mapping;
-	}
-
-	if (job_render->include_images) {
-		if (job_info->image_mapping)
-			ev_image_mapping_free (job_info->image_mapping);
-		job_info->image_mapping = job_render->image_mapping;
-	}
-
-	if (job_render->include_text) {
-		if (job_info->text_mapping)
-			gdk_region_destroy (job_info->text_mapping);
-		job_info->text_mapping = job_render->text_mapping;
-	}
-
+	job_info->points_set = FALSE;
 	if (job_render->include_selection) {
 		if (job_info->selection) {
 			cairo_surface_destroy (job_info->selection);
@@ -455,6 +478,44 @@ copy_job_to_job_info (EvJobRender   *job_render,
 		job_info->points_set = TRUE;
 	}
 
+	if (job_info->job) {
+		g_signal_handlers_disconnect_by_func (job_info->job,
+						      G_CALLBACK (job_page_ready_cb),
+						      pixbuf_cache);
+	}
+
+	job_info->page_ready = TRUE;
+}
+
+static void
+copy_job_to_job_info (EvJobRender   *job_render,
+		      CacheJobInfo  *job_info,
+		      EvPixbufCache *pixbuf_cache)
+{
+	if (job_render->include_links) {
+		if (job_info->link_mapping)
+			ev_link_mapping_free (job_info->link_mapping);
+		job_info->link_mapping = job_render->link_mapping;
+	}
+
+	if (job_render->include_images) {
+		if (job_info->image_mapping)
+			ev_image_mapping_free (job_info->image_mapping);
+		job_info->image_mapping = job_render->image_mapping;
+	}
+
+	if (job_render->include_forms) {
+		if (job_info->form_field_mapping)
+			ev_form_field_mapping_free (job_info->form_field_mapping);
+		job_info->form_field_mapping = job_render->form_field_mapping;
+	}
+
+	if (job_render->include_text) {
+		if (job_info->text_mapping)
+			gdk_region_destroy (job_info->text_mapping);
+		job_info->text_mapping = job_render->text_mapping;
+	}
+
 	if (job_info->job) {
 		g_signal_handlers_disconnect_by_func (job_info->job,
 						      G_CALLBACK (job_finished_cb),
@@ -532,33 +593,26 @@ get_selection_colors (GtkWidget *widget, GdkColor **text, GdkColor **base)
 }
 
 static void
-add_job_if_needed (EvPixbufCache *pixbuf_cache,
-		   CacheJobInfo  *job_info,
-		   EvPageCache   *page_cache,
-		   gint           page,
-		   gint           rotation,
-		   gfloat         scale,
-		   EvJobPriority  priority)
+add_job (EvPixbufCache *pixbuf_cache,
+	 CacheJobInfo  *job_info,
+	 EvPageCache   *page_cache,
+	 GdkRegion     *region,
+	 gint           width,
+	 gint           height,
+	 gint           page,
+	 gint           rotation,
+	 gfloat         scale,
+	 EvJobPriority  priority)
 {
 	gboolean include_links = FALSE;
 	gboolean include_text = FALSE;
 	gboolean include_selection = FALSE;
-	gboolean include_images = FALSE;
-	int width, height;
+	gboolean include_images = TRUE;
+	gboolean include_forms = FALSE;
 	GdkColor *text, *base;
 
-	if (job_info->job)
-		return;
-
-	ev_page_cache_get_size (page_cache, page, rotation,
-				scale, &width, &height);
-
-	if (job_info->surface &&
-	    cairo_image_surface_get_width (job_info->surface) == width &&
-	    cairo_image_surface_get_height (job_info->surface) == height)
-		return;
-
-	/* make a new job now */
+	job_info->page_ready = FALSE;
+	
 	if (job_info->rc == NULL) {
 		job_info->rc = ev_render_context_new (rotation, page, scale);
 	} else {
@@ -567,11 +621,17 @@ add_job_if_needed (EvPixbufCache *pixbuf_cache,
 		ev_render_context_set_scale (job_info->rc, scale);
 	}
 
+	if (job_info->region)
+		gdk_region_destroy (job_info->region);
+	job_info->region = region ? gdk_region_copy (region) : NULL;
+
 	/* Figure out what else we need for this job */
 	if (job_info->link_mapping == NULL)
 		include_links = TRUE;
 	if (job_info->image_mapping == NULL)
 		include_images = TRUE;
+	if (job_info->form_field_mapping == NULL)
+		include_forms = TRUE;
 	if (job_info->text_mapping == NULL)
 		include_text = TRUE;
 	if (new_selection_surface_needed (pixbuf_cache, job_info, page, scale)) {
@@ -587,14 +647,46 @@ add_job_if_needed (EvPixbufCache *pixbuf_cache,
 					   width, height,
 					   &(job_info->target_points),
 					   text, base,
+					   include_forms,
 					   include_links,
 					   include_images,
 					   include_text,
 					   include_selection);
 	ev_job_queue_add_job (job_info->job, priority);
-	g_signal_connect (job_info->job, "finished", G_CALLBACK (job_finished_cb), pixbuf_cache);
+	g_signal_connect (G_OBJECT (job_info->job), "page-ready",
+			  G_CALLBACK (job_page_ready_cb),
+			  pixbuf_cache);
+	g_signal_connect (G_OBJECT (job_info->job), "finished",
+			  G_CALLBACK (job_finished_cb),
+			  pixbuf_cache);
 }
 
+static void
+add_job_if_needed (EvPixbufCache *pixbuf_cache,
+		   CacheJobInfo  *job_info,
+		   EvPageCache   *page_cache,
+		   gint           page,
+		   gint           rotation,
+		   gfloat         scale,
+		   EvJobPriority  priority)
+{
+	gint width, height;
+
+	if (job_info->job)
+		return;
+
+	ev_page_cache_get_size (page_cache, page, rotation,
+				scale, &width, &height);
+
+	if (job_info->surface &&
+	    cairo_image_surface_get_width (job_info->surface) == width &&
+	    cairo_image_surface_get_height (job_info->surface) == height)
+		return;
+
+	add_job (pixbuf_cache, job_info, page_cache, NULL,
+		 width, height, page, rotation, scale,
+		 priority);
+}
 
 static void
 ev_pixbuf_cache_add_jobs_if_needed (EvPixbufCache *pixbuf_cache,
@@ -681,11 +773,14 @@ ev_pixbuf_cache_get_surface (EvPixbufCache *pixbuf_cache,
 	if (job_info == NULL)
 		return NULL;
 
+	if (job_info->page_ready)
+		return job_info->surface;
+	
 	/* We don't need to wait for the idle to handle the callback */
 	if (job_info->job &&
-	    EV_JOB (job_info->job)->finished) {
-		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
-		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
+	    EV_JOB_RENDER (job_info->job)->page_ready) {
+		copy_job_page_and_selection_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
+		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0, job_info->region);
 	}
 
 	return job_info->surface;
@@ -705,7 +800,6 @@ ev_pixbuf_cache_get_link_mapping (EvPixbufCache *pixbuf_cache,
 	if (job_info->job &&
 	    EV_JOB (job_info->job)->finished) {
 		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
-		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
 	}
 
 	return job_info->link_mapping;
@@ -728,12 +822,33 @@ ev_pixbuf_cache_get_image_mapping (EvPixbufCache *pixbuf_cache,
 	if (job_info->job &&
 	    EV_JOB (job_info->job)->finished) {
 		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
-		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
 	}
 
 	return job_info->image_mapping;
 }
 
+GList *
+ev_pixbuf_cache_get_form_field_mapping (EvPixbufCache *pixbuf_cache,
+				  	gint 	       page)
+{
+	CacheJobInfo *job_info;
+
+	if (!EV_IS_DOCUMENT_FORMS (pixbuf_cache->document))
+		return NULL;
+	
+	job_info = find_job_cache (pixbuf_cache, page);
+	if (job_info == NULL)
+		return NULL;
+
+	/* We don't need to wait for the idle to handle the callback */
+	if (job_info->job &&
+	   EV_JOB (job_info->job)->finished) {
+		copy_job_to_job_info (EV_JOB_RENDER(job_info->job), job_info, pixbuf_cache);
+	}
+	
+	return job_info->form_field_mapping;
+}
+
 static gboolean
 new_selection_surface_needed (EvPixbufCache *pixbuf_cache,
 			     CacheJobInfo  *job_info,
@@ -792,7 +907,6 @@ ev_pixbuf_cache_get_text_mapping (EvPixbufCache *pixbuf_cache,
 	if (job_info->job &&
 	    EV_JOB (job_info->job)->finished) {
 		copy_job_to_job_info (EV_JOB_RENDER (job_info->job), job_info, pixbuf_cache);
-		g_signal_emit (pixbuf_cache, signals[JOB_FINISHED], 0);
 	}
 	
 	return job_info->text_mapping;
@@ -1102,3 +1216,28 @@ ev_pixbuf_cache_get_selection_list (EvPixbufCache *pixbuf_cache)
 	return retval;
 }
 
+void           
+ev_pixbuf_cache_reload_page (EvPixbufCache *pixbuf_cache,
+			     GdkRegion     *region,
+			     gint           page,
+			     gint           rotation,
+			     gdouble        scale)
+{
+	CacheJobInfo *job_info;
+        EvPageCache *page_cache;
+        gint width, height;
+
+	job_info = find_job_cache (pixbuf_cache, page);
+	if (job_info == NULL)
+		return;
+	
+        page_cache = ev_page_cache_get (pixbuf_cache->document);
+	ev_page_cache_get_size (page_cache, page, rotation, scale,
+				&width, &height);
+
+        add_job (pixbuf_cache, job_info, page_cache, region,
+		 width, height, page, rotation, scale,
+		 EV_JOB_PRIORITY_HIGH);
+}
+
+
